"""
PyTorch utilities.

| Copyright 2017-2020, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
import logging
import itertools
import multiprocessing

import numpy as np
from PIL import Image

from eta.core.config import Config
import eta.core.learning as etal
import eta.core.utils as etau

import fiftyone as fo
import fiftyone.core.labels as fol
import fiftyone.core.media as fomm
import fiftyone.core.models as fom
import fiftyone.core.utils as fou
import fiftyone.core.validation as fov

fou.ensure_torch()
import torch
import torchvision
from torchvision.transforms import functional as F
from torch.utils.data import Dataset, DataLoader


logger = logging.getLogger(__name__)


def apply_torch_image_model(
    samples,
    model,
    label_field,
    confidence_thresh=None,
    num_workers=None,
    batch_size=None,
):
    """Applies the :class:`fiftyone.core.models.Model` to the image samples in
    the collection.

    This method performs the same operation as
    :func:`fiftyone.core.models.apply_model` except that the images are loaded
    using a ``torch.utils.data.DataLoader``.

    The ``model`` must implement the
    :class:`fiftyone.core.models.TorchModelMixin` mixin.

    Args:
        samples: a :class:`fiftyone.core.collections.SampleCollection`
        model: a :class:`fiftyone.core.models.Model`
        label_field: the name (or prefix) of the field in which to store the
            model predictions
        confidence_thresh (None): an optional confidence threshold to apply to
            any classifications/objects generated by the model
        num_workers (None): the number of workers for the
            ``torch.utils.data.DataLoader`` to use
        batch_size (None): an optional batch size to use
    """
    if samples.media_type != fomm.IMAGE:
        raise ValueError("This method only supports image samples")

    if not isinstance(model, fom.TorchModelMixin):
        raise ValueError(
            "Model must implement the %s mixin" % fom.TorchModelMixin
        )

    batch_size = _parse_batch_size(batch_size, model)
    samples_loader = fou.iter_batches(samples, batch_size)
    data_loader = _make_data_loader(samples, model, batch_size, num_workers)

    with fou.ProgressBar(samples) as pb:
        with fou.SetAttributes(model, preprocess=False):
            with model:
                for sample_batch, imgs in zip(samples_loader, data_loader):
                    labels_batch = model.predict_all(imgs)

                    for sample, labels in zip(sample_batch, labels_batch):
                        sample.add_labels(
                            labels,
                            label_field,
                            confidence_thresh=confidence_thresh,
                        )

                    pb.set_iteration(pb.iteration + len(imgs))


def compute_torch_image_embeddings(
    samples, model, embeddings_field=None, num_workers=None, batch_size=None
):
    """Computes embeddings for the samples in the collection using the given
    :class:`fiftyone.core.models.Model`.

    This method performs the same operation as
    :func:`fiftyone.core.models.compute_embeddings` except that the images are
    loaded using a ``torch.utils.data.DataLoader``.

    The ``model`` must implement the
    :class:`fiftyone.core.models.TorchModelMixin` mixin, and it must expose
    embeddings, i.e., :meth:`fiftyone.core.models.Model.has_embeddings` must
    return ``True``.

    If an ``embeddings_field`` is provided, the embeddings are saved to the
    samples; otherwise, the embeddings are returned in-memory.

    Args:
        samples: a :class:`fiftyone.core.collections.SampleCollection`
        model: a :class:`fiftyone.core.models.Model`
        embeddings_field (None): the name of a field in which to store the
            embeddings
        num_workers (None): the number of workers for the
            ``torch.utils.data.DataLoader`` to use
        batch_size (None): an optional batch size to use

    Returns:
        ``None``, if an ``embeddings_field`` is provided; otherwise, a numpy
        array whose first dimension is ``len(samples)`` containing the
        embeddings
    """
    if samples.media_type != fomm.IMAGE:
        raise ValueError("This method only supports image samples")

    if not isinstance(model, fom.TorchModelMixin):
        raise ValueError(
            "Model must implement the %s mixin" % fom.TorchModelMixin
        )

    if not model.has_embeddings:
        raise ValueError(
            "Model does not expose embeddings (model.has_embeddings = %s)"
            % model.has_embeddings
        )

    batch_size = _parse_batch_size(batch_size, model)
    data_loader = _make_data_loader(samples, model, batch_size, num_workers)

    if embeddings_field:
        samples_loader = fou.iter_batches(samples, batch_size)
    else:
        samples_loader = itertools.repeat(None)

    embeddings = []

    with fou.ProgressBar(samples) as pb:
        with fou.SetAttributes(model, preprocess=False):
            with model:
                for sample_batch, imgs in zip(samples_loader, data_loader):
                    embeddings_batch = model.embed_all(imgs)

                    if embeddings_field:
                        for sample, embedding in zip(
                            sample_batch, embeddings_batch
                        ):
                            sample[embeddings_field] = embedding
                            sample.save()
                    else:
                        embeddings.append(embeddings_batch)

                    pb.set_iteration(pb.iteration + len(imgs))

    if embeddings_field:
        return None

    return np.concatenate(embeddings)


def compute_torch_image_patch_embeddings(
    samples,
    model,
    patches_field,
    embeddings_field=None,
    force_square=False,
    alpha=None,
    num_workers=None,
    batch_size=None,
):
    """Computes embeddings for the image patches defined by ``patches_field``
    of the samples in the collection using the given :class:`Model`.

    This method performs the same operation as
    :func:`fiftyone.core.models.compute_patch_embeddings` except that the
    patches are loaded using a ``torch.utils.data.DataLoader``.

    The ``model`` must implement the
    :class:`fiftyone.core.models.TorchModelMixin` mixin, and it must expose
    embeddings, i.e., :meth:`fiftyone.core.models.Model.has_embeddings` must
    return ``True``.

    If an ``embeddings_field`` is provided, the embeddings are saved to the
    samples; otherwise, the embeddings are returned in-memory.

    Args:
        samples: a :class:`fiftyone.core.collections.SampleCollection`
        model: a :class:`Model`
        patches_field: a :class:`fiftyone.core.labels.Detection`,
            :class:`fiftyone.core.labels.Detections`,
            :class:`fiftyone.core.labels.Polyline`, or
            :class:`fiftyone.core.labels.Polylines` field defining the image
            patches in each sample to embed
        embeddings_field (None): the name of a field in which to store the
            embeddings
        force_square (False): whether to minimally manipulate the patch
            bounding boxes into squares prior to extraction
        alpha (None): an optional expansion/contraction to apply to the patches
            before extracting them, in ``[-1, \infty)``. If provided, the
            length and width of the box are expanded (or contracted, when
            ``alpha < 0``) by ``(100 * alpha)%``. For example, set
            ``alpha = 1.1`` to expand the boxes by 10%, and set ``alpha = 0.9``
            to contract the boxes by 10%
        num_workers (None): the number of workers for the
            ``torch.utils.data.DataLoader`` to use
        batch_size (None): an optional batch size to use

    Returns:
        ``None``, if an ``embeddings_field`` is provided; otherwise, a dict
        mapping sample IDs to patch embeddings
    """
    if samples.media_type != fomm.IMAGE:
        raise ValueError("This method only supports image samples")

    if not isinstance(model, fom.TorchModelMixin):
        raise ValueError(
            "Model must implement the %s mixin" % fom.TorchModelMixin
        )

    if not model.has_embeddings:
        raise ValueError(
            "Model does not expose embeddings (model.has_embeddings = %s)"
            % model.has_embeddings
        )

    allowed_types = (
        fol.Detection,
        fol.Detections,
        fol.Polyline,
        fol.Polylines,
    )
    fov.validate_collection_label_fields(samples, patches_field, allowed_types)

    batch_size = _parse_batch_size(batch_size, model)
    data_loader = _make_patch_data_loader(
        samples, model, patches_field, force_square, alpha, num_workers,
    )

    embeddings_dict = {}

    with fou.ProgressBar(samples) as pb:
        with fou.SetAttributes(model, preprocess=False):
            with model:
                for sample, patches in pb(zip(samples, data_loader)):
                    if patches is None:
                        continue

                    embeddings = []
                    for patches_batch in fou.iter_slices(patches, batch_size):
                        embeddings_batch = model.embed_all(patches_batch)
                        embeddings.append(embeddings_batch)

                    embeddings = np.concatenate(embeddings)

                    if embeddings_field:
                        sample[embeddings_field] = embeddings
                        sample.save()
                    else:
                        embeddings_dict[sample.id] = embeddings

    return embeddings_dict if not embeddings_field else None


def _parse_batch_size(batch_size, model):
    if batch_size is None:
        batch_size = fo.config.default_batch_size or 1

    if batch_size > 1 and model.ragged_batches:
        logger.warning("Model does not support batching")
        batch_size = 1

    return batch_size


def _make_data_loader(samples, model, batch_size, num_workers):
    if num_workers is None:
        num_workers = recommend_num_workers()

    image_paths = []
    for sample in samples.select_fields():
        image_paths.append(sample.filepath)

    dataset = TorchImageDataset(
        image_paths, transform=model.transforms, force_rgb=True
    )

    if model.ragged_batches:
        kwargs = dict(collate_fn=lambda batch: batch)  # return list
    else:
        kwargs = {}

    return DataLoader(
        dataset, batch_size=batch_size, num_workers=num_workers, **kwargs
    )


def _make_patch_data_loader(
    samples, model, patches_field, force_square, alpha, num_workers
):
    if num_workers is None:
        num_workers = recommend_num_workers()

    image_paths = []
    detections = []
    for sample in samples.select_fields(patches_field):
        patches = _parse_patches(sample, patches_field)
        image_paths.append(sample.filepath)
        detections.append(patches)

    dataset = TorchImagePatchesDataset(
        image_paths,
        detections,
        model.transforms,
        ragged_batches=model.ragged_batches,
        force_rgb=True,
        force_square=force_square,
        alpha=alpha,
    )

    return DataLoader(
        dataset,
        batch_size=1,
        num_workers=num_workers,
        collate_fn=lambda batch: batch[0],  # return patches directly
    )


def _parse_patches(sample, patches_field):
    label = sample[patches_field]

    if isinstance(label, fol.Detections):
        return label

    if isinstance(label, fol.Detection):
        return fol.Detections(detections=[label])

    if isinstance(label, fol.Polyline):
        return fol.Detections(detections=[label.to_detection()])

    if isinstance(label, fol.Polylines):
        return label.to_detections()

    return None


class TorchEmbeddingsMixin(fom.EmbeddingsMixin):
    """Mixin for Torch models that can generate embeddings.

    Args:
        model: the Torch model, a ``torch.nn.Module``
        layer_name (None): the name of the embeddings layer whose output to
            save, or ``None`` if this model instance should not expose
            embeddings. Prepend ``"<"`` to save the input tensor instead
    """

    def __init__(self, model, layer_name=None):
        if layer_name is not None:
            embeddings_layer = SaveLayerTensor(model, layer_name)
        else:
            embeddings_layer = None

        self._embeddings_layer = embeddings_layer

    @property
    def has_embeddings(self):
        return self._embeddings_layer is not None

    def embed(self, arg):
        if isinstance(arg, torch.Tensor):
            args = arg.unsqueeze(0)
        else:
            args = [arg]

        self._predict_all(args)
        return self.get_embeddings()[0]

    def embed_all(self, args):
        self._predict_all(args)
        return self.get_embeddings()

    def get_embeddings(self):
        if not self.has_embeddings:
            raise ValueError("This model instance does not expose embeddings")

        embeddings = self._embeddings_layer.tensor.detach().cpu().numpy()
        return embeddings.astype(float, copy=False)

    def _predict_all(self, args):
        """Applies a forward pass to the given iterable of data and returns
        the raw model output with no processing applied.

        Args:
            args: an iterable of data. See :meth:`predict_all` for details

        Returns:
            the raw output of the model
        """
        raise NotImplementedError("subclasses must implement _predict_all()")


class TorchImageModelConfig(Config):
    """Configuration for running a :class:`TorchImageModel`.

    Args:
        entrypoint_fcn: a fully-qualified function string like
            ``"torchvision.models.inception_v3"`` specifying the entrypoint
            function that loads the model
        entrypoint_args (None): a dictionary of arguments for
            ``entrypoint_fcn``
        output_processor_cls: a string like
            ``"fifytone.utils.torch.ClassifierOutputProcessor"`` specifying the
            :class:`fifytone.utils.torch.OutputProcessor` to use
        output_processor_args (None): a dictionary of arguments for
            ``output_processor_cls(class_labels, **kwargs)``
        labels_string (None): a comma-separated list of the class names for the
            model
        labels_path (None): the path to the labels map for the model
        image_min_size (None): resize the input images during preprocessing, if
            necessary, so that the image dimensions are at least this
            ``(width, height)``
        image_min_dim (None): resize input images during preprocessing, if
            necessary, so that the smaller image dimension is at least this
            value
        image_size (None): a ``(width, height)`` to which to resize the input
            images during preprocessing
        image_dim (None): resize the smaller input dimension to this value
            during preprocessing
        image_mean (None): a 3-array of mean values in ``[0, 1]`` for
            preprocessing the input images
        image_std (None): a 3-array of std values in ``[0, 1]`` for
            preprocessing the input images
            inputs that are lists of Tensors
        embeddings_layer (None): the name of a layer whose output to expose as
            embeddings. Prepend ``"<"`` to save the input tensor instead
        use_half_precision (None): whether to use half precision (only
            supported when using GPU)
        cudnn_benchmark (None): a value to use for
            ``torch.backends.cudnn.benchmark`` while the model is running
    """

    def __init__(self, d):
        self.entrypoint_fcn = self.parse_string(d, "entrypoint_fcn")
        self.entrypoint_args = self.parse_dict(
            d, "entrypoint_args", default=None
        )
        self.output_processor_cls = self.parse_string(
            d, "output_processor_cls"
        )
        self.output_processor_args = self.parse_dict(
            d, "output_processor_args", default=None
        )
        self.labels_string = self.parse_string(
            d, "labels_string", default=None
        )
        self.labels_path = self.parse_string(d, "labels_path", default=None)
        self.image_min_size = self.parse_array(
            d, "image_min_size", default=None
        )
        self.image_min_dim = self.parse_number(
            d, "image_min_dim", default=None
        )
        self.image_size = self.parse_array(d, "image_size", default=None)
        self.image_dim = self.parse_number(d, "image_dim", default=None)
        self.image_mean = self.parse_array(d, "image_mean", default=None)
        self.image_std = self.parse_array(d, "image_std", default=None)
        self.embeddings_layer = self.parse_string(
            d, "embeddings_layer", default=None
        )
        self.use_half_precision = self.parse_bool(
            d, "use_half_precision", default=None
        )
        self.cudnn_benchmark = self.parse_bool(
            d, "cudnn_benchmark", default=None
        )


class TorchImageModel(TorchEmbeddingsMixin, fom.TorchModelMixin, fom.Model):
    """Wrapper for evaluating a Torch model on images.

    This wrapper assumes that ``config.entrypoint_fcn`` returns a
    ``torch.nn.Module`` whose ``__call__()`` method directly accepts Torch
    tensors (NCHW) as input.

    Args:
        config: an :class:`TorchImageModelConfig`
    """

    def __init__(self, config):
        self.config = config

        # Get class labels
        self._class_labels = self._get_class_labels(config)
        self._num_classes = len(self._class_labels)

        # Build transforms
        ragged_batches, transforms = self._build_transforms(config)
        self._ragged_batches = ragged_batches
        self._transforms = transforms
        self._preprocess = True

        # Load model
        self._using_gpu = torch.cuda.is_available()
        self._device = torch.device("cuda:0" if self._using_gpu else "cpu")
        self._using_half_precision = (
            self.config.use_half_precision is True
        ) and self._using_gpu
        self._model = self._load_model(config)
        self._no_grad = None
        self._benchmark_orig = None

        # Initialize embeddings saver (if necessary)
        TorchEmbeddingsMixin.__init__(
            self, self._model, layer_name=self.config.embeddings_layer
        )

        # Build output processor
        self._output_processor = self._build_output_processor(config)

    def __enter__(self):
        if self.config.cudnn_benchmark is not None:
            self._benchmark_orig = torch.backends.cudnn.benchmark
            torch.backends.cudnn.benchmark = self.config.cudnn_benchmark

        self._no_grad = torch.no_grad()
        self._no_grad.__enter__()
        return self

    def __exit__(self, *args):
        if self.config.cudnn_benchmark is not None:
            torch.backends.cudnn.benchmark = self._benchmark_orig
            self._benchmark_orig = None

        self._no_grad.__exit__(*args)
        self._no_grad = None

    @property
    def ragged_batches(self):
        """True/False whether :meth:`transforms` may return tensors of
        different sizes. If True, then passing ragged lists of images to
        :meth:`predict_all` is not allowed.
        """
        return self._ragged_batches

    @property
    def transforms(self):
        """The ``torchvision.transforms`` function that will/must be applied to
        each input before prediction.
        """
        return self._transforms

    @property
    def preprocess(self):
        """Whether to apply preprocessing transforms during inference."""
        return self._preprocess

    @preprocess.setter
    def preprocess(self, value):
        self._preprocess = value

    @property
    def using_gpu(self):
        """Whether the model is using GPU."""
        return self._using_gpu

    @property
    def device(self):
        """The ``torch.device`` that the model is using."""
        return self._device

    @property
    def using_half_precision(self):
        """Whether the model is using half precision."""
        return self._using_half_precision

    @property
    def num_classes(self):
        """The number of classes for the model."""
        return self._num_classes

    @property
    def class_labels(self):
        """The list of class labels for the model."""
        return self._class_labels

    def predict(self, img):
        """Peforms prediction on the given image.

        Args:
            img: the image to process, which can be any of the following:

                - A PIL image
                - A uint8 numpy array (HWC)
                - A Torch tensor (CWH)

        Returns:
            a :class:`fiftyone.core.labels.Label` instance or dict of
            :class:`fiftyone.core.labels.Label` instances containing the
            predictions
        """
        if isinstance(img, torch.Tensor):
            imgs = img.unsqueeze(0)
        else:
            imgs = [img]

        return self._predict_all(imgs)[0]

    def predict_all(self, imgs):
        """Peforms prediction on the given batch of images.

        Args:
            imgs: the batch of images to process, which can be any of the
                following:

                - A list of PIL images
                - A list of uint8 numpy arrays (HWC)
                - A list of Torch tensors (CHW)
                - A uint8 numpy tensor (NHWC)
                - A Torch tensor (NCHW)

        Returns:
            a list of :class:`fiftyone.core.labels.Label` instances or a list
            of dicts of :class:`fiftyone.core.labels.Label` instances
            containing the predictions
        """
        return self._predict_all(imgs)

    def _predict_all(self, imgs):
        if self._preprocess:
            imgs = [self._transforms(img) for img in imgs]

        if isinstance(imgs, (list, tuple)):
            imgs = torch.stack(imgs)

        height, width = imgs.size()[-2:]
        frame_size = (width, height)

        if self._using_gpu:
            imgs = imgs.cuda()

        if self._using_half_precision:
            imgs = imgs.half()

        output = self._model(imgs)
        return self._output_processor(output, frame_size)

    def _get_class_labels(self, config):
        if config.labels_string:
            return config.labels_string.split(",")

        if not config.labels_path:
            raise ValueError(
                "Either `labels_string` or `labels_path` must be specified"
            )

        labels_path = fou.fill_patterns(config.labels_path)
        labels_map = etal.load_labels_map(labels_path)
        return etal.get_class_labels(labels_map)

    def _build_transforms(self, config):
        ragged_batches = True
        transforms = [ToPILImage()]

        if config.image_min_size:
            transforms.append(MinResize(config.image_min_size))
        elif config.image_min_dim:
            transforms.append(MinResize(config.image_min_dim))
        elif config.image_size:
            ragged_batches = False
            transforms.append(torchvision.transforms.Resize(config.image_size))
        elif config.image_dim:
            transforms.append(torchvision.transforms.Resize(config.image_dim))

        # Converts PIL/numpy (HWC) to Torch tensor (CHW) in [0, 1]
        transforms.append(torchvision.transforms.ToTensor())

        if config.image_mean or config.image_std:
            if not config.image_mean or not config.image_std:
                raise ValueError(
                    "Both `image_mean` and `image_std` must be provided"
                )

            transforms.append(
                torchvision.transforms.Normalize(
                    config.image_mean, config.image_std
                )
            )

        transforms = torchvision.transforms.Compose(transforms)
        return ragged_batches, transforms

    def _load_model(self, config):
        self._download_model(config)

        model = self._load_network(config)

        model = model.to(self._device)
        if self._using_half_precision:
            model = model.half()

        self._load_state_dict(model, config)

        model.train(False)

        return model

    def _download_model(self, config):
        pass

    def _load_network(self, config):
        entrypoint = etau.get_function(config.entrypoint_fcn)
        kwargs = config.entrypoint_args or {}
        return entrypoint(**kwargs)

    def _load_state_dict(self, model, config):
        pass

    def _build_output_processor(self, config):
        output_processor_cls = etau.get_class(config.output_processor_cls)
        kwargs = config.output_processor_args or {}
        return output_processor_cls(self._class_labels, **kwargs)


class ToPILImage(object):
    """Transform that converts a tensor or ndarray to a PIL image, while also
    allowing PIL images to passthrough.
    """

    def __call__(self, img):
        if isinstance(img, Image.Image):
            return img

        return F.to_pil_image(img)


class MinResize(object):
    """Transform that resizes the PIL image or torch Tensor, if necessary, so
    that its minimum dimensions are at least the specified size.

    Args:
        min_output_size: Desired minimum output dimensions. Can either be a
            ``(min_height, min_width)`` tuple or a single ``min_dim``
        interpolation (None): Optional interpolation mode. Passed directly to
            ``torchvision.transforms.functional.resize``
    """

    def __init__(self, min_output_size, interpolation=None):
        if isinstance(min_output_size, int):
            min_output_size = (min_output_size, min_output_size)

        self.min_output_size = min_output_size
        self.interpolation = interpolation

        self._kwargs = {}
        if interpolation is not None:
            self._kwargs["interpolation"] = interpolation

    def __call__(self, pil_image_or_tensor):
        if isinstance(pil_image_or_tensor, torch.Tensor):
            h, w = list(pil_image_or_tensor.size())[-2:]
        else:
            w, h = pil_image_or_tensor.size

        minh, minw = self.min_output_size

        if h >= minh and w >= minw:
            return pil_image_or_tensor

        alpha = max(minh / h, minw / w)
        size = (int(round(alpha * h)), int(round(alpha * w)))
        return F.resize(pil_image_or_tensor, size, **self._kwargs)


class SaveLayerTensor(object):
    """Callback that saves the input/output tensor of the specified layer of a
    Torch model during each ``forward()`` call.

    Args:
        model: the Torch model, a ``torch.nn.Module``
        layer_name: the name of the layer whose output to save. Prepend ``"<"``
            to save the input tensor instead
    """

    def __init__(self, model, layer_name):
        self._tensor = None
        self._save_input = None
        self._setup(model, layer_name)

    def __call__(self, module, module_in, module_out):
        self._tensor = module_in[0] if self._save_input else module_out

    @property
    def tensor(self):
        """The tensor saved from the last ``forward()`` call."""
        return self._tensor

    def _setup(self, model, layer_name):
        if layer_name.startswith("<"):
            self._save_input = True
            layer_name = layer_name[1:]
        else:
            self._save_input = False

        _layer = None
        for name, layer in model.named_modules():
            if name == layer_name:
                _layer = layer

        if _layer is None:
            raise ValueError("No layer found with name %s" % layer_name)

        _layer.register_forward_hook(self)


class OutputProcessor(object):
    """Interface for processing the outputs of Torch models."""

    def __call__(self, output, frame_size):
        """Parses the model output.

        Args:
            output: the model output for the batch of predictions
            frame_size: the ``(width, height)`` of the frames in the batch

        Returns:
            a list of :class:`fiftyone.core.labels.Label` instances
        """
        raise NotImplementedError("subclass must implement __call__")


class ClassifierOutputProcessor(OutputProcessor):
    """Output processor for single label classifiers.

    Args:
        class_labels: the list of class labels for the model
        confidence_thresh (None): an optional confidence threshold to apply
            when deciding whether to keep predictions
    """

    def __init__(self, class_labels, confidence_thresh=None):
        self.class_labels = class_labels
        self.confidence_thresh = confidence_thresh

    def __call__(self, output, _):
        """Parses the model output.

        Args:
            output: either a ``FloatTensor[N, M]`` containing the logits for
                ``N`` images and ``M`` classes, or a dict with a ``"logits"``
                key containing the logits

        Returns:
            a list of :class:`fiftyone.core.labels.Classification` instances
        """
        if isinstance(output, dict):
            output = output["logits"]

        logits = output.detach().cpu().numpy()

        predictions = np.argmax(logits, axis=1)
        odds = np.exp(logits)
        odds /= np.sum(odds, axis=1, keepdims=True)
        scores = np.max(odds, axis=1)

        preds = []
        for prediction, score in zip(predictions, scores):
            if (
                self.confidence_thresh is not None
                and score < self.confidence_thresh
            ):
                classification = None
            else:
                classification = fol.Classification(
                    label=self.class_labels[prediction], confidence=score,
                )

            preds.append(classification)

        return preds


class DetectorOutputProcessor(OutputProcessor):
    """Output processor for object detectors.

    Args:
        class_labels: the list of class labels for the model
        confidence_thresh (None): an optional confidence threshold to apply
            when deciding whether to keep predictions
    """

    def __init__(self, class_labels, confidence_thresh=None):
        self.class_labels = class_labels
        self.confidence_thresh = confidence_thresh

    def __call__(self, output, frame_size):
        """Parses the model output.

        Args:
            output: a batch of predictions ``output = List[Dict[Tensor]]``,
                where each dict has the following keys:

                -   boxes (``FloatTensor[N, 4]``): the predicted boxes in
                    ``[x1, y1, x2, y2]`` format (absolute coordinates)
                -   labels (``Int64Tensor[N]``): the predicted labels
                -   scores (``Tensor[N]``): the scores for each prediction

            frame_size: the ``(width, height)`` of the frames in the batch

        Returns:
            a list of :class:`fiftyone.core.labels.Detections` instances
        """
        return [self._parse_output(o, frame_size) for o in output]

    def _parse_output(self, output, frame_size):
        width, height = frame_size

        boxes = output["boxes"].detach().cpu().numpy()
        labels = output["labels"].detach().cpu().numpy()
        scores = output["scores"].detach().cpu().numpy()

        detections = []
        for box, label, score in zip(boxes, labels, scores):
            if (
                self.confidence_thresh is not None
                and score < self.confidence_thresh
            ):
                continue

            x1, y1, x2, y2 = box
            bounding_box = [
                x1 / width,
                y1 / height,
                (x2 - x1) / width,
                (y2 - y1) / height,
            ]

            detections.append(
                fol.Detection(
                    label=self.class_labels[label],
                    bounding_box=bounding_box,
                    confidence=score,
                )
            )

        return fol.Detections(detections=detections)


class InstanceSegmenterOutputProcessor(OutputProcessor):
    """Output processor for instance segementers.

    Args:
        class_labels: the list of class labels for the model
        confidence_thresh (None): an optional confidence threshold to apply
            when deciding whether to keep predictions
        mask_thresh (0.5): a threshold to use to convert soft masks to binary
            masks
    """

    def __init__(self, class_labels, confidence_thresh=None, mask_thresh=0.5):
        self.class_labels = class_labels
        self.confidence_thresh = confidence_thresh
        self.mask_thresh = mask_thresh

    def __call__(self, output, frame_size):
        """Parses the model output.

        Args:
            output: a batch of predictions ``output = List[Dict[Tensor]]``,
                where each dict has the following keys:

                -   boxes (``FloatTensor[N, 4]``): the predicted boxes in
                    ``[x1, y1, x2, y2]`` format (absolute coordinates)
                -   labels (``Int64Tensor[N]``): the predicted labels
                -   scores (``Tensor[N]``): the scores for each prediction
                -   masks (``FloatTensor[N, 1, H, W]``): the predicted masks
                    for each instance, in ``[0, 1]``

            frame_size: the ``(width, height)`` of the frames in the batch

        Returns:
            a list of :class:`fiftyone.core.labels.Detections` instances
        """
        return [self._parse_output(o, frame_size) for o in output]

    def _parse_output(self, output, frame_size):
        width, height = frame_size

        boxes = output["boxes"].detach().cpu().numpy()
        labels = output["labels"].detach().cpu().numpy()
        scores = output["scores"].detach().cpu().numpy()
        masks = output["masks"].detach().cpu().numpy()

        detections = []
        for box, label, score, soft_mask in zip(boxes, labels, scores, masks):
            if (
                self.confidence_thresh is not None
                and score < self.confidence_thresh
            ):
                continue

            x1, y1, x2, y2 = box
            bounding_box = [
                x1 / width,
                y1 / height,
                (x2 - x1) / width,
                (y2 - y1) / height,
            ]

            soft_mask = np.squeeze(soft_mask, axis=0)[
                int(round(y1)) : int(round(y2)),
                int(round(x1)) : int(round(x2)),
            ]
            mask = soft_mask > self.mask_thresh

            detections.append(
                fol.Detection(
                    label=self.class_labels[label],
                    bounding_box=bounding_box,
                    mask=mask,
                    confidence=score,
                )
            )

        return fol.Detections(detections=detections)


class KeypointDetectorOutputProcessor(OutputProcessor):
    """Output processor for keypoint detection models.

    Args:
        class_labels: the list of class labels for the model
        edges (None): an optional list of list of vertices specifying polyline
            connections between keypoints to draw
        confidence_thresh (None): an optional confidence threshold to apply
            when deciding whether to keep predictions
    """

    def __init__(self, class_labels, edges=None, confidence_thresh=None):
        self.class_labels = class_labels
        self.edges = edges
        self.confidence_thresh = confidence_thresh

    def __call__(self, output, frame_size):
        """Parses the model output.

        Args:
            output: a batch of predictions ``output = List[Dict[Tensor]]``,
                where each dict has the following keys:

                -   boxes (``FloatTensor[N, 4]``): the predicted boxes in
                    ``[x1, y1, x2, y2]`` format (absolute coordinates)
                -   labels (``Int64Tensor[N]``): the predicted labels
                -   scores (``Tensor[N]``): the scores for each prediction
                -   keypoints (``FloatTensor[N, K, ...]``): the predicted
                    keypoints for each instance in ``[x, y, ...]`` format

            frame_size: the ``(width, height)`` of the frames in the batch

        Returns:
            a list of :class:`fiftyone.core.labels.Label` dicts
        """
        return [self._parse_output(o, frame_size) for o in output]

    def _parse_output(self, output, frame_size):
        width, height = frame_size

        boxes = output["boxes"].detach().cpu().numpy()
        labels = output["labels"].detach().cpu().numpy()
        scores = output["scores"].detach().cpu().numpy()
        keypoints = output["keypoints"].detach().cpu().numpy()

        _detections = []
        _keypoints = []
        _polylines = []
        for box, label, score, kpts in zip(boxes, labels, scores, keypoints):
            if (
                self.confidence_thresh is not None
                and score < self.confidence_thresh
            ):
                continue

            x1, y1, x2, y2 = box
            bounding_box = [
                x1 / width,
                y1 / height,
                (x2 - x1) / width,
                (y2 - y1) / height,
            ]

            points = [(p[0] / width, p[1] / height) for p in kpts]

            _detections.append(
                fol.Detection(
                    label=self.class_labels[label],
                    bounding_box=bounding_box,
                    confidence=score,
                )
            )

            _keypoints.append(
                fol.Keypoint(
                    label=self.class_labels[label],
                    points=points,
                    confidence=score,
                )
            )

            if self.edges is not None:
                _polylines.append(
                    fol.Polyline(
                        points=[[points[v] for v in e] for e in self.edges],
                        confidence=score,
                        closed=False,
                        filled=False,
                    )
                )

        label = {
            "detections": fol.Detections(detections=_detections),
            "keypoints": fol.Keypoints(keypoints=_keypoints),
        }

        if self.edges is not None:
            label["polylines"] = fol.Polylines(polylines=_polylines)

        return label


class SegmenterOutputProcessor(OutputProcessor):
    """Output processor for semantic segementers.

    Args:
        class_labels: the list of class labels for the model
    """

    def __init__(self, class_labels):
        self.class_labels = class_labels

    def __call__(self, output, _):
        """Parses the model output.

        Args:
            output: a batch of predictions ``output = Dict[Tensor]``,
                where the dict has the following keys:

                -   out (``FloatTensor[N, M, H, W]``): the segmentation map
                    probabilities for the ``N`` images across the ``M`` classes

        Returns:
            a list of :class:`fiftyone.core.labels.Segmentation` instances
        """
        probs = output["out"].detach().cpu().numpy()
        masks = probs.argmax(axis=1)
        return [fol.Segmentation(mask=mask) for mask in masks]


def recommend_num_workers():
    """Recommend a number of workers for running a
    ``torch.utils.data.DataLoader``.

    Returns:
        the recommended ``num_workers``
    """
    if not torch.cuda.is_available() and torch.__version__.startswith("1.7"):
        # There is a parallelism bug in torch==1.7 on CPU that prevents us from
        # using `num_workers > 0`
        # https://stackoverflow.com/q/64772335
        return 0

    try:
        return multiprocessing.cpu_count() // 2
    except:
        return 4


class TorchImageDataset(Dataset):
    """A ``torch.utils.data.Dataset`` of images.

    Instances of this class emit images, or ``(image, sample_id)`` pairs if
    ``sample_ids`` are provided.

    Args:
        image_paths: an iterable of image paths
        sample_ids (None): an iterable of :class:`fiftyone.core.sample.Sample`
            IDs corresponding to each image
        transform (None): an optional transform to apply to the images
        force_rgb (False): whether to force convert the images to RGB
    """

    def __init__(
        self, image_paths, sample_ids=None, transform=None, force_rgb=False
    ):
        self.image_paths = list(image_paths)
        self.sample_ids = list(sample_ids) if sample_ids else None
        self.transform = transform
        self.force_rgb = force_rgb

    def __len__(self):
        return len(self.image_paths)

    def __getitem__(self, idx):
        img = Image.open(self.image_paths[idx])

        if self.force_rgb:
            img = img.convert("RGB")

        if self.transform:
            img = self.transform(img)

        if self.has_sample_ids:
            # pylint: disable=unsubscriptable-object
            return img, self.sample_ids[idx]

        return img

    @property
    def has_sample_ids(self):
        """Whether this dataset has sample IDs."""
        return self.sample_ids is not None


class TorchImageClassificationDataset(Dataset):
    """A ``torch.utils.data.Dataset`` for image classification.

    Instances of this dataset emit images and their associated targets, either
    directly as ``(image, target)`` pairs or as ``(image, target, sample_id)``
    pairs if ``sample_ids`` are provided.

    Args:
        image_paths: an iterable of image paths
        targets: an iterable of targets
        sample_ids (None): an iterable of :class:`fiftyone.core.sample.Sample`
            IDs corresponding to each image
        transform (None): an optional transform to apply to the images
        force_rgb (False): whether to force convert the images to RGB
    """

    def __init__(
        self,
        image_paths,
        targets,
        sample_ids=None,
        transform=None,
        force_rgb=False,
    ):
        self.image_paths = list(image_paths)
        self.targets = list(targets)
        self.sample_ids = list(sample_ids) if sample_ids else None
        self.transform = transform
        self.force_rgb = force_rgb

    def __len__(self):
        return len(self.image_paths)

    def __getitem__(self, idx):
        img = Image.open(self.image_paths[idx])
        target = self.targets[idx]

        if self.force_rgb:
            img = img.convert("RGB")

        if self.transform:
            img = self.transform(img)

        if self.has_sample_ids:
            # pylint: disable=unsubscriptable-object
            return img, target, self.sample_ids[idx]

        return img, target

    @property
    def has_sample_ids(self):
        """Whether this dataset has sample IDs."""
        return self.sample_ids is not None


class TorchImagePatchesDataset(Dataset):
    """A ``torch.utils.data.Dataset`` of image patch tensors extracted from a
    list of images.

    Instances of this class emit Torch tensors containing the stacked
    (along axis 0) patches from each image, or ``(patch_tensor, sample_id)``
    pairs if ``sample_ids`` are provided.

    The provided ``transform`` must ensure that all image patches are resized
    to the same shape and formatted as torch Tensors so that they can be
    stacked.

    Args:
        image_paths: an iterable of image paths
        detections: an iterable of :class:`fiftyone.core.labels.Detections`
            instances specifying the image patch(es) to extract from each
            image
        transform: a torchvision transform to apply to each image patch
        sample_ids (None): an iterable of :class:`fiftyone.core.sample.Sample`
            IDs corresponding to each image
        ragged_batches (False): whether the provided ``transform`` may return
            tensors of different dimensions
        force_rgb (False): whether to force convert the images to RGB
        force_square (False): whether to minimally manipulate the patch
            bounding boxes into squares prior to extraction
        alpha (None): an optional expansion/contraction to apply to the patches
            before extracting them, in ``[-1, \infty)``. If provided, the
            length and width of the box are expanded (or contracted, when
            ``alpha < 0``) by ``(100 * alpha)%``. For example, set
            ``alpha = 1.1`` to expand the boxes by 10%, and set ``alpha = 0.9``
            to contract the boxes by 10%
    """

    def __init__(
        self,
        image_paths,
        detections,
        transform,
        sample_ids=None,
        ragged_batches=False,
        force_rgb=False,
        force_square=False,
        alpha=None,
    ):
        self.image_paths = list(image_paths)
        self.detections = list(detections)
        self.transform = transform
        self.sample_ids = list(sample_ids) if sample_ids else None
        self.ragged_batches = ragged_batches
        self.force_rgb = force_rgb
        self.force_square = force_square
        self.alpha = alpha

    def __len__(self):
        return len(self.image_paths)

    def __getitem__(self, idx):
        detections = self.detections[idx]
        if detections is None or not detections.detections:
            img_patches = None
        else:
            image_path = self.image_paths[idx]
            img_patches = self._extract_patches(image_path, detections)

        if self.has_sample_ids:
            # pylint: disable=unsubscriptable-object
            return img_patches, self.sample_ids[idx]

        return img_patches

    @property
    def has_sample_ids(self):
        """Whether this dataset has sample IDs."""
        return self.sample_ids is not None

    def _extract_patches(self, image_path, detections):
        img = Image.open(image_path)

        if self.force_rgb:
            img = img.convert("RGB")

        # @todo avoid PIL <-> numpy casts
        img = np.array(img)

        img_patches = []
        for detection in detections.detections:
            dobj = detection.to_detected_object()

            bbox = dobj.bounding_box
            if self.alpha is not None:
                bbox = bbox.pad_relative(self.alpha)

            img_patch = bbox.extract_from(img, force_square=self.force_square)
            img_patch = Image.fromarray(img_patch)
            img_patch = self.transform(img_patch)
            img_patches.append(img_patch)

        if self.ragged_batches:
            return img_patches

        return torch.stack(img_patches, dim=0)


def from_image_classification_dir_tree(dataset_dir):
    """Creates a ``torch.utils.data.Dataset`` for the given image
    classification dataset directory tree.

    The directory should have the following format::

        <dataset_dir>/
            <classA>/
                <image1>.<ext>
                <image2>.<ext>
                ...
            <classB>/
                <image1>.<ext>
                <image2>.<ext>
                ...

    Args:
        dataset_dir: the dataset directory

    Returns:
        a ``torchvision.datasets.ImageFolder``
    """
    return torchvision.datasets.ImageFolder(dataset_dir)
